(***********************************************************************)
(*                                                                     *)
(*                           FaCiLe                                    *)
(*                                                                     *)
(*            Nicolas Barnier, Pascal Brisset, LOG, ENAC/CENA          *)
(*                                                                     *)
(***********************************************************************)
(* $Id: golf.ml,v 1.14 2000/11/17 08:51:11 brisset Exp $ *)

(*
  A Golf Tournament (from http://www.icparc.ic.ac.uk/~cg6/conjunto.html)

  There are 32 golfers, who play individually but in groups of 4,
  called foursomes. The tournament is organized in weeks.  Each week
  a new set of foursomes has to be computed such that each person only
  golfs with the same person once. So if two golfers have played each
  other in any previous week they should not play each other in the
  coming weeks. The question is: "how many weeks can we ensure this
  before players start to play each other a second time ?"
*)

open Facile
open Easy

(* Two modes using the Global Cardinality Constraint or the Sort constraint *)
type mode = Gcc | Sort
let mode_of = function
    "gcc" -> Gcc | "sort" -> Sort | _ -> failwith "Unknown mode";;

let go nb_golfers size_group nb_weeks mode =
  let nb_groups = nb_golfers / size_group in

  (* An array of nb_weeks*nb_golfers decision variables to choose the group
     (in 0..nb_groups-1) of every golfer every week *)
  let vars =
    Array.init
      nb_weeks
      ~f:(fun _ -> Fd.array nb_golfers 0 (nb_groups-1)) in

  (* Constraints *)
  (* For each week, exactly 4 golfers in each group: *)
  begin
    match mode with
      Gcc ->
	(* Using a Global Cardinality Constraint *)
  	let cards_values = Array.init nb_groups ~f:(fun i -> (Fd.int 4, i)) in
  	for w = 0 to nb_weeks - 1 do
	  Cstr.post (Gcc.cstr vars.(w) cards_values)
  	done
    | Sort ->
	(* Using a Sorting constraint: For each week the sorted array of
	   groups is equal to [|0;0;0;0;1;1;1;1;2;2;2;2;....|] *)
	let sorted =
	  Array.init nb_golfers ~f:(fun i -> Fd.int (i / size_group)) in
	for j = 0 to nb_weeks - 1 do
	  Cstr.post (Sorting.cstr vars.(j) sorted)
  	done
  end;

  (* Two golfers do not play in the same group more than once *)
  for g1 = 0 to nb_golfers - 1 do (* for each pair of golfers *)
    for g2 = g1+1 to nb_golfers - 1 do
      let g1_with_g2 =
	Array.init nb_weeks ~f:(fun w -> Arith.e2fd (fd2e vars.(w).(g1) =~~ fd2e vars.(w).(g2))) in
      Cstr.post (Arith.sum_fd g1_with_g2 <=~ i2e 1)
    done
  done;

  (* Breaking the symmetries
     0 always in the first group, 1 in a group less than 1, ...
     First week (0) a priori chosen
     *)
  for w = 0 to nb_weeks - 1 do
    for g = 0 to nb_groups - 1 do
      Cstr.post (fd2e vars.(w).(g) <=~ i2e g)
    done
  done;
  for g = 0 to nb_golfers - 1 do 
    Cstr.post (fd2e vars.(0).(g) =~ i2e (g / size_group))
  done;

  (* Seach goal: Choose the groups for the first golfer, then for second
     one, ... *)
  let goal =
    Goals.forto 0 (nb_golfers-1)
      (fun g -> Goals.forto 0 (nb_weeks-1)
	  (fun w -> Goals.indomain vars.(w).(g))) in

  (* Solving *)
  if Goals.solve goal then begin
    for w = 0 to nb_weeks - 1 do
      for g = 0 to nb_groups do
      	for p = 0 to nb_golfers - 1 do
	  if Fd.min vars.(w).(p) = g then Printf.printf "%2d " p
	done;
	print_string " ";
      done;
      print_newline ()
    done
  end else
    prerr_endline "No solution";;

let _ =
  if Array.length Sys.argv < 3 then
    prerr_endline "Usage: golf <nb_weeks> <mode> (gcc or sort)\nNo solutions for nb_weeks > 9"
  else
    let nb_weeks = int_of_string Sys.argv.(1)
    and mode = mode_of Sys.argv.(2) in
    go 32 4 nb_weeks mode;;
  
