(***********************************************************************)
(*                                                                     *)
(*                           FaCiLe                                    *)
(*                                                                     *)
(*            Nicolas Barnier, Pascal Brisset, LOG, ENAC/CENA          *)
(*                                                                     *)
(***********************************************************************)
(* $Id: facile.mli,v 1.37 2001/01/15 13:16:20 barnier Exp $ *)

(* Module [Facile]: end-user interface to the FaCiLe library *)

module Domain :
  sig
    type t
    val empty : t
    val create : int list -> t
    val interval : int -> int -> t
    val int : t
    val boolean : t
    val is_empty : t -> bool
    val size : t -> int
    val min : t -> int
    val max : t -> int
    val iter : (int -> unit) -> t -> unit
    val member : int -> t -> bool
    val values : t -> int list
    val fprint : out_channel -> t -> unit
    val sprint : t -> string
    val included : t -> t -> bool
    val remove : int -> t -> t
    val add : int -> t -> t
    val remove_up : int -> t -> t
    val remove_low : int -> t -> t
    val remove_closed_inter : int -> int -> t -> t
    val intersection : t -> t -> t
    val union : t -> t -> t
    val difference : t -> t -> t
    val remove_min : t -> t
    val minus : t -> t
    val plus : t -> int -> t
    val first_geq_value : t -> int -> int
    val first_leq_value : t -> int -> int
    val choose : (int -> int -> bool) -> t -> int
  end
module Stak :
  sig
    type level
    exception Level_not_found of level
    val size : unit -> int
    val level : unit -> level
    val cut : level -> unit
    val cut_bottom : level -> unit
    type 'a ref
    val ref : 'a -> 'a ref
    val set : 'a ref -> 'a -> unit
    val get : 'a ref -> 'a
    exception Fail of string
    val fail : string -> 'a
  end
module Cstr :
  sig
    exception DontKnow
    type priority
    val immediate : priority
    val normal : priority
    val later : priority
    type t
    val id : t -> int
    val name : t -> string
    val fprint : out_channel -> t -> unit
    val create :
      ?name:string ->
	?fprint:(out_channel -> unit) ->
  	  ?priority:priority ->
	    ?init:(unit -> unit) ->
	      ?check:(unit -> bool) ->
		?not:(unit -> t) ->
		    (unit -> bool) ->
		      (t -> unit) ->
			t
    val post : t -> unit
    val init : t -> unit
    val one : t
    val zero : t
    val active_store : unit -> t list
  end
module Var :
  sig
    module Attr :
      sig
        type t
	type event
        val dom : t -> Domain.t
        val on_refine : event
        val on_subst : event
        val on_min : event
        val on_max : event
        val fprint : out_channel -> t -> unit
        val size : t -> int
        val min : t -> int
        val max : t -> int
        val values : t -> int list
        val iter : t -> (int -> unit) -> unit
        val member : t -> int -> bool
        val id : t -> int
	val constraints_number : t -> int
      end
    type concrete_fd = Unk of Attr.t | Val of int
    module Fd :
      sig
        type t
        val create : Domain.t -> t
        val int : int -> t
        val interval : int -> int -> t
  	val array : int -> int -> int -> t array
  	val value : t -> concrete_fd
        val subst : t -> int -> unit
        val refine : t -> Domain.t -> unit
        val unify : t -> int -> unit
        val fprint : out_channel -> t -> unit
        val fprint_array : out_channel -> t array -> unit
        val size : t -> int
        val min : t -> int
        val max : t -> int
        val values : t -> int list
        val is_var : t -> bool
        val iter : t -> (int -> unit) -> unit
        val member : t -> int -> bool
        val id : t -> int
        val compare : t -> t -> int
        val equal : t -> t -> bool
      end
    val delay : Attr.event list -> Fd.t -> Cstr.t -> unit
  end
module Reify :
  sig
    val boolean : ?delay_on_negation:bool -> Cstr.t -> Var.Fd.t
    val cstr : ?delay_on_negation:bool -> Cstr.t -> Var.Fd.t -> Cstr.t
    val (||~~) : Cstr.t -> Cstr.t -> Cstr.t
    val (&&~~) : Cstr.t -> Cstr.t -> Cstr.t
    val (<=>~~) : Cstr.t -> Cstr.t -> Cstr.t
    val not : Cstr.t -> Cstr.t
    val (=>~~) : Cstr.t -> Cstr.t -> Cstr.t
  end
module Alldiff :
  sig
    type algo =
        Lazy
      | Bin_matching of Var.Attr.event
    val cstr : ?algo:algo -> Var.Fd.t array -> Cstr.t
  end
module Goals :
  sig
    type t
    val create : ?name:string -> (unit -> unit) -> t
    val make : ?name:string -> ('a -> t option) -> 'a -> t
    val make_rec : ?name:string -> (t -> t option) -> t
    val fail : t
    val success : t
    val ( &&~ ) : t -> t -> t
    val ( ||~ ) : t -> t -> t
    val once : t -> t
    val solve : ?control:(int -> unit) -> t -> bool
    val unify : Var.Fd.t -> int -> t
    val indomain : Var.Fd.t -> t
    val instantiate : (Domain.t -> int) -> Var.Fd.t -> t
    val dichotomic : Var.Fd.t -> t
    val forto : int -> int -> (int -> t) -> t
    val fordownto : int -> int -> (int -> t) -> t
    module GlArray :
      sig
        val iter_h : ('a array -> int) -> ('a -> t) -> 'a array -> t
        val iter_hi :
          ('a array -> int) -> (int -> 'a -> t) -> 'a array -> t
        val iter : ('a -> t) -> 'a array -> t
        val iteri : (int -> 'a -> t) -> 'a array -> t
        val iter2 : ('a -> 'b -> t) -> 'a array -> 'b array -> t
        val labeling : Var.Fd.t array -> t
        val choose_index :
            (Var.Attr.t -> Var.Attr.t -> bool) -> Var.Fd.t array -> int
        val not_instantiated_fd : Var.Fd.t array -> int
      end
    module GlList :
      sig
        val iter : ('a -> t) -> 'a list -> t
        val labeling : Var.Fd.t list -> t
        val member : Var.Fd.t -> int list -> t
        val iter_h : ('a list -> 'a * 'a list) -> ('a -> t) -> 'a list -> t
      end
  end
module Sorting :
  sig
    val sort : Var.Fd.t array -> Var.Fd.t array
    val sortp :
      Var.Fd.t array -> Var.Fd.t array * Var.Fd.t array
    val cstr :
      Var.Fd.t array ->
      ?p:Var.Fd.t array option -> Var.Fd.t array -> Cstr.t
  end
module Arith :
  sig
    type t
    val fprint : out_channel -> t -> unit
    val eval : t -> int
    val min_of_expr : t -> int
    val max_of_expr : t -> int
    val i2e : int -> t
    val fd2e : Var.Fd.t -> t
    val ( +~ ) : t -> t -> t
    val ( *~ ) : t -> t -> t
    val ( -~ ) : t -> t -> t
    val ( /~ ) : t -> t -> t
    val ( **~ ) : t -> int -> t
    val ( %~ ) : t -> t -> t
    val abs : t -> t
    val sum : t array -> t
    val sum_fd : Var.Fd.t array -> t
    val scalprod : int array -> t array -> t
    val scalprod_fd : int array -> Var.Fd.t array -> t
    val prod : t array -> t
    val prod_fd : Var.Fd.t array -> t
    val ( <=~ ) : t -> t -> Cstr.t
    val ( <~ ) : t -> t -> Cstr.t
    val ( >~ ) : t -> t -> Cstr.t
    val ( =~ ) : t -> t -> Cstr.t
    val ( <>~ ) : t -> t -> Cstr.t
    val ( >=~ ) : t -> t -> Cstr.t
    val e2fd : t -> Var.Fd.t
    val even : t -> Cstr.t
    val odd : t -> Cstr.t
    val ( <=~~ ) : t -> t -> t
    val ( <~~ ) : t -> t -> t
    val ( >~~ ) : t -> t -> t
    val ( =~~ ) : t -> t -> t
    val ( <>~~ ) : t -> t -> t
    val ( >=~~ ) : t -> t -> t
  end
module Interval :
  sig
    val is_member : Var.Fd.t -> int -> int -> Var.Fd.t
    val cstr : Var.Fd.t -> int -> int -> Var.Fd.t -> Cstr.t
  end
module FdArray :
  sig
    val min : Var.Fd.t array -> Var.Fd.t
    val min_cstr : Var.Fd.t array -> Var.Fd.t -> Cstr.t
    val max : Var.Fd.t array -> Var.Fd.t
    val max_cstr : Var.Fd.t array -> Var.Fd.t -> Cstr.t
    val get : Var.Fd.t array -> Var.Fd.t -> Var.Fd.t
    val get_cstr : Var.Fd.t array -> Var.Fd.t -> Var.Fd.t -> Cstr.t
  end
module Gcc :
  sig
    type level = Basic | Medium | High
    val cstr :
      ?level:level ->
      Var.Fd.t array ->
      (Var.Fd.t * int) array -> Cstr.t
  end
module Opti :
  sig
    val minimize :
      Goals.t -> Var.Fd.t ->
	?control:(int -> unit) ->
	  ?step:int -> (int -> 'a) -> 'a option
  end
module Easy :
  sig
    val i2e : int -> Arith.t
    val fd2e : Var.Fd.t -> Arith.t
    val ( +~ ) : Arith.t -> Arith.t -> Arith.t
    val ( *~ ) : Arith.t -> Arith.t -> Arith.t
    val ( -~ ) : Arith.t -> Arith.t -> Arith.t
    val ( /~ ) : Arith.t -> Arith.t -> Arith.t
    val ( **~ ) : Arith.t -> int -> Arith.t
    val ( %~ ) : Arith.t -> Arith.t -> Arith.t
    val ( <=~ ) : Arith.t -> Arith.t -> Cstr.t
    val ( <~ ) : Arith.t -> Arith.t -> Cstr.t
    val ( >~ ) : Arith.t -> Arith.t -> Cstr.t
    val ( =~ ) : Arith.t -> Arith.t -> Cstr.t
    val ( <>~ ) : Arith.t -> Arith.t -> Cstr.t
    val ( >=~ ) : Arith.t -> Arith.t -> Cstr.t
    val ( <=~~ ) : Arith.t -> Arith.t -> Arith.t
    val ( <~~ ) : Arith.t -> Arith.t -> Arith.t
    val ( >~~ ) : Arith.t -> Arith.t -> Arith.t
    val ( =~~ ) : Arith.t -> Arith.t -> Arith.t
    val ( <>~~ ) : Arith.t -> Arith.t -> Arith.t
    val ( >=~~ ) : Arith.t -> Arith.t -> Arith.t
    val (&&~~) : Cstr.t -> Cstr.t -> Cstr.t
    val (||~~) : Cstr.t -> Cstr.t -> Cstr.t
    val (=>~~) : Cstr.t -> Cstr.t -> Cstr.t
    val (<=>~~) : Cstr.t -> Cstr.t -> Cstr.t
    val ( &&~ ) : Goals.t -> Goals.t -> Goals.t
    val ( ||~ ) : Goals.t -> Goals.t -> Goals.t
    module Fd :
      sig
        type t = Var.Fd.t
        val create : Domain.t -> t
        val interval : int -> int -> t
        val array : int -> int -> int -> t array
        val int : int -> t
	val value : t -> Var.concrete_fd
        val subst : t -> int -> unit
        val refine : t -> Domain.t -> unit
        val unify : t -> int -> unit
        val fprint : out_channel -> t -> unit
        val fprint_array : out_channel -> t array -> unit
        val size : t -> int
        val min : t -> int
        val max : t -> int
        val values : t -> int list
        val is_var : t -> bool
        val iter : t -> (int -> unit) -> unit
        val member : t -> int -> bool
        val id : t -> int
        val compare : t -> t -> int
        val equal : t -> t -> bool
      end
    type concrete_fd = Var.concrete_fd = Unk of Var.Attr.t | Val of int
  end
  
