(***********************************************************************)
(*                                                                     *)
(*                           FaCiLe                                    *)
(*                                                                     *)
(*            Nicolas Barnier, Pascal Brisset, LOG, ENAC/CENA          *)
(*                                                                     *)
(***********************************************************************)
(* $Id: fcl_arith.mli,v 1.21 2001/01/29 10:46:56 brisset Exp $ *)

(* Module [Arith]: arithmetic expressions over variables of type [Var.Fd.t] *)

(*** Overview *)
(*q This module provides functions and operators to build arithmetic
   expressions and state (in|dis)equation constraints over them.
*)  

(*** Basic *)
type t
(* Type of arithmetic expressions over variables of type [Var.Fd.t] and
   integers. *)
val fprint : out_channel -> t -> unit
(* [fprint chan e] prints expression [e] on channel [chan]. *)
val eval : t -> int
(* [eval e] returns the integer numerical value of a fully instantiated
   expression [e]. Raises [Invalid_argument] if [e] is not instantiated. *)
val min_of_expr : t -> int
(* [min_of_expr e] returns the minimal possible value of expression [e]. *)
val max_of_expr : t -> int
(* [max_of_expr e] returns the maximal possible value of expression [e]. *)

(* Conversions *)
val i2e : int -> t
(* [i2e n] returns an expression which evaluates to [n]. *)
val fd2e : Fcl_var.Fd.t -> t
(* [fd2e v] returns an expression which evaluates to [n] when [v] is
   instantiated and [Var.Fd.value v] evaluates to [Val n]. *)
val e2fd : t -> Fcl_var.Fd.t
(* [e2fd e] creates and returns a new variable [v] and posts the constraint
   [fd2e v =~ e]. *)

(*** Construction of arithmetic expressions *)
(*q
  The arithmetic operators check whether any integer overflow (i.e. the
   result of an arithmetic operation over integers is less than [min_int]
   or greater than [max_int]) occurs during constraints intern computations
   and raise an assert failure only when using the byte code library
  [facile.cma]. *)
val (+~) : t -> t -> t
(* Addition over expressions. *)
val ( *~) : t -> t -> t
(* Multiplication over expressions. *)
val (-~) : t -> t -> t
(* Substraction over expressions. *)
val (/~) : t -> t -> t
(* Division over expressions. A failure occurs whenever the second
   argument evaluates to 0. *)
val ( **~) : t -> int -> t
(* Exponentiation of an expression to an integer value. *)
val (%~) : t -> t -> t
(* Modulo over expressions. *)
val abs : t -> t
(* Absolute value over expressions. *)
val sum : t array -> t
(* Sum of all the elements of an array of expressions. *)
val sum_fd : Fcl_var.Fd.t array -> t
(* Sum of all the elements of an array of variables. *)
val scalprod : int array -> t array -> t
(* Scalar product of an array of integers and an array of expressions.
   Raises [Invalid_argument] if the arrays have not the same length. *)
val scalprod_fd : int array -> Fcl_var.Fd.t array -> t
(* Scalar product of an array of integers and an array of variables.
   Raises [Invalid_argument] if the arrays have not the same length. *)
val prod : t array -> t
(* Product of all the elements of an array of expressions. *)
val prod_fd : Fcl_var.Fd.t array -> t
(* Product of all the elements of an array of variables. *)

(*** Arithmetic constraints over expressions *)
val (<=~) : t -> t -> Fcl_cstr.t
(* [e1 <=~ e2] returns the constraint ensuring that the expression [e1] is
   less or equal than expression [e2]. *)
val (<~)  : t -> t -> Fcl_cstr.t
(* [e1 <~ e2] returns the constraint ensuring that the expression [e1] is
   strictly less than expression [e2]. *)
val (>~)  : t -> t -> Fcl_cstr.t
(* [e1 >~ e2] returns the constraint ensuring that the expression [e1] is
   strictly greater than expression [e2]. *)
val (=~)  : t -> t -> Fcl_cstr.t
(* [e1 =~ e2] returns the constraint ensuring that the expression [e1] is
   equal to expression [e2]. *)
val (<>~)  : t -> t -> Fcl_cstr.t
(* [e1 =~ e2] returns the constraint ensuring that the expression [e1] is
   different from expression [e2]. *)
val (>=~) : t -> t -> Fcl_cstr.t
(* [e1 >=~ e2] returns the constraint ensuring that the expression [e1] is
   greater or equal than expression [e2]. *)
val even : t -> Fcl_cstr.t
(* [even e] returns the constraint ensuring that that expression [e] is even. *)
val odd : t -> Fcl_cstr.t
(* [odd e] returns the constraint ensuring that that expression [e] is odd. *)

(*** Reification *)
(*q 
   Reification of the arithmetic constraint which is replaced by an expression equal to
   a boolean variable instantiated to [1] if the constraint is satisfied and
   to [0] if it is violated. *)
val (<=~~) : t -> t -> t
(* Reified less or equal. [e1 <=~~ e2] is equivalent to [fd2e (Reify.boolean (e1 <=~ e2))]. *)
val (<~~)  : t -> t -> t
(* Reified strictly less. *)
val (>~~)  : t -> t -> t
(* Reified strictly greater. *)
val (>=~~) : t -> t -> t
(* Reified greater or equal. *)
val (=~~)  : t -> t -> t
(* Reified equality. *)
val (<>~~)  : t -> t -> t
(* Reified different. *)
