(***********************************************************************)
(*                                                                     *)
(*                           FaCiLe                                    *)
(*                                                                     *)
(*            Nicolas Barnier, Pascal Brisset, LOG, ENAC/CENA          *)
(*                                                                     *)
(***********************************************************************)
(* $Id: fcl_goals.ml,v 1.19 2001/01/29 14:23:41 brisset Exp $ *)

open Fcl_var

class virtual t = Fcl_stak.gl

let save = Fcl_stak.save

exception Success

let goal_stack = ref [];;

let reset () = Fcl_stak.reset (); Fcl_cstr.reset_queue ();;

let while_true = fun _ -> ()


let solve = fun ?(control = while_true) goal ->
  Fcl_misc.protect "Goals.solve"
    (fun () ->
      Fcl_stak.reset ();
      let backtracks = ref 0 in
      goal_stack := [goal];
      let rec or_loop () =
    	try
	  Fcl_cstr.reset_queue ();
	  let rec and_loop () =
            match !goal_stack with
	      [] -> raise Success
    	    | goal::goals ->
            	goal_stack := goals;
            	begin
		  match goal#call with
		    None -> ()
		  | Some g -> goal_stack := g :: !goal_stack
		end;
            	and_loop ()
	  in
	  and_loop ()
    	with
	  Fcl_stak.Fail s ->
            Fcl_debug.call 'g' (fun f ->  Printf.fprintf f "fail %s\n" s);
            goal_stack := Fcl_stak.backtrack ();
            incr backtracks;
            control !backtracks ;
            or_loop () in
      try
    	or_loop ()
      with
    	Fcl_stak.Empty_stack -> false
      | Success -> control !backtracks; true);;

class make_rec name f = object (self)
  inherit t
  method name = name
  method call = f self
end
let make_rec ?(name = "make_rec") f = new make_rec name f

let make ?(name = "make") f a = make_rec ~name:name (fun _self -> f a)
let create ?(name = "create") f = make_rec ~name:name (fun _self -> f (); None)

let (&&~) g1 g2 =
  make ~name:"&&~" (fun () -> goal_stack := g2:: !goal_stack; Some g1) ()
let (||~) g1 g2 =
  make ~name:"||~" (fun () -> ignore(Fcl_stak.save (g2:: !goal_stack)); Some g1) ()
  
let fail = create ~name:"fail" (fun () -> Fcl_stak.fail "fail");;
let success = create ~name:"success" (fun () -> ());;

let indomain var =
  make_rec ~name:"indomain"
    (fun self ->
      match Fd.value var with
      Val _ -> None
    | Unk var_ -> 
    	let mini = Fcl_domain.min (Attr.dom var_) in
      	Some (
	  create (fun () -> Fd.subst var mini) ||~
	  create (fun () -> Fd.refine var (Fcl_domain.remove_min (Attr.dom var_))) &&~ self));;

let dichotomic var =
  make_rec ~name:"dichotomic"
    (fun self ->
      match Fd.value var with
      	Val _ -> None
      | Unk var_ -> 
      	  let d = Attr.dom var_ in
      	  let mini = Fcl_domain.min d and maxi = Fcl_domain.max d in
      	  let middle = (maxi + mini) / 2 in
	  Some (
	  (create (fun () -> Fd.refine var (Fcl_domain.remove_up middle d))
         ||~
	   create (fun () -> Fd.refine var (Fcl_domain.remove_low (middle+1) d)))
	    &&~
	  self))

let instantiate choose var =
  make_rec ~name:"instantiate"
    (fun self ->
      match Fd.value var with
      Val _ -> None
    | Unk var_ -> 
    	let x = choose (Attr.dom var_) in
      	Some (
	  create (fun () -> Fd.subst var x) ||~
	  create (fun () -> Fd.refine var (Fcl_domain.remove x (Attr.dom var_))) &&~ self));;


let once goal =
  let l = ref (Fcl_stak.level ()) in
  create (fun () -> l := (Fcl_stak.level ())) &&~
  goal &&~
  create (fun () -> Fcl_stak.cut !l)

let forto min max g =
  let rec la i =
    if i <= max
    then Some (g i &&~ make la (i+1))
    else None
  in
  make ~name:"forto" la min

let fordownto max min g =
  forto min max (fun i -> g (max - i + min));;

module GlArray = struct
  let iter_hi select lab_one a =
    make_rec
      (fun self ->
      	try
	  let i = select a in
	  Some (lab_one i a.(i) &&~ self)
      	with
	  Not_found -> None)

  let iter_h select lab_one a = iter_hi select (fun _ -> lab_one) a

  let iteri lab_one a =
    let size = Array.length a in
    let rec la i =
      if i < size
      then Some (lab_one i a.(i) &&~ make la (i+1))
      else None
    in
    make la 0

  let iter f a = iteri (fun _i -> f) a

  let iter2 goal a b=
    let size = Array.length a in
    if size  <> Array.length b then
      invalid_arg "Goals.GlArray.iter2";
    let rec la i =
      if i < size
      then Some (goal a.(i) b.(i) &&~ make la (i+1))
      else None
    in
    make la 0

  let choose_index order tab =
    let n = Array.length tab in

    (* Recherche de la premiere variable libre *)
    let rec first_unbound i = 
      if i < n then
      	match Fd.value tab.(i) with
	  Unk attr -> i, attr
	| Val _ -> first_unbound (i+1)
      else
	raise Not_found in
    
    let b, attr = first_unbound 0 in
    let best = ref b and attr_best = ref attr in
    (* Recherche de la meilleure variable pour le critere *)
    for i = b+1 to n - 1 do
      match Fd.value tab.(i) with
	Unk tabi ->
	  if order tabi !attr_best then begin
	    best := i; attr_best := tabi
	  end
      |	Val _ -> ()
    done;
    !best

  let labeling (a : Fcl_var.Fd.t array) = iter indomain a

  exception Return of int
  let not_instantiated_fd fds =
    try
      for i = 0 to Array.length fds - 1 do
      	if Fd.is_var fds.(i) then raise (Return i)
      done;
      raise Not_found
    with
      Return i -> i
end

module GlList = struct
  let iter fgoal l =
    let rec cl l =
      match  l with
      	[] -> None
      | x::xs -> Some (fgoal x &&~ make cl xs) in
    make cl l

  let iter_disj fgoal l =
    let rec cl l =
      match  l with
      	[] -> Some fail
      |	[x] -> Some (fgoal x) (* To avoid a choice point on fail *)
      | x::xs -> Some (fgoal x ||~ make cl xs) in
    make cl l

  let member v l =
    iter_disj (fun x -> create (fun () -> Fd.unify v x)) l

  let iter_h select fgoal l =
    let rec clh l =
      match l with
	[] -> None
      |	_ ->
        let x,xs = select l in
        Some (fgoal x &&~ make clh xs)
    in
    make clh l

  let labeling (l : Fcl_var.Fd.t list) = iter indomain l
end

let save = Fcl_stak.save

let unify v x = create (fun () -> Fd.unify v x)
